
from colibris.authorization import ANY_PERMISSION

from __packagename__ import views
from __packagename__.constants import ROLE_ADMIN, ROLE_REGULAR

#
# Routes example:
#
# ROUTES = [
#     ('GET',    '/users/me',   views.get_me,       ANY_PERMISSION),
#     ('GET',    '/users/{id}', views.get_user,     {ROLE_ADMIN}),
#     ('GET',    '/users',      views.list_users,   {ROLE_ADMIN}),
#     ('POST',   '/users',      views.add_user,     {ROLE_ADMIN}),
#     ('PATCH',  '/users/{id}', views.update_user,  {ROLE_ADMIN}),
#     ('DELETE', '/users/{id}', views.delete_user,  {ROLE_ADMIN}),
#
#     ('GET',    '/my_resource/{id}', views.get_my_resource, {ROLE_ADMIN, ROLE_REGULAR}),
# ]
#




import importlib
import inspect
import os
import re
import sys

from dotenv import load_dotenv

from colibris import utils

from colibris.conf import defaultsettings
from colibris.conf import lazysettings
from colibris.conf import schemas as settings_schemas


_settings_store = {}


def _is_setting_name(name):
    # only consider public members that are all in capitals
    return re.match('^[A-Z][A-Z0-9_]*$', name)


def _override_setting(name, value):
    # do we have the setting corresponding to the given name?
    if name in _settings_store:
        _settings_store[name] = value
        return

    # try dictionary with items
    parts = name.split('_')
    for i in range(len(parts) - 1):
        dname = '_'.join(parts[:i + 1])
        d = _settings_store.get(dname)
        if not isinstance(d, dict):
            continue

        dkey = '_'.join(parts[i + 1:])
        d[dkey.lower()] = value  # dictionary keys are lower-case, by convention

        break

    else:
        # lastly, we simply add the new setting to the module
        _settings_store[name] = value


def _setup_project_package():
    # autodetect project package from main script path
    main_script = sys.argv[0]
    project_package_name = None
    if main_script.endswith('manage.py'):  # using manage.py
        main_script = os.path.realpath(main_script)
        project_package_name = os.path.basename(os.path.dirname(main_script))
        project_package_name = re.sub('[^a-z0-9_]', '', project_package_name).lower()

    else:  # using a setuptools console script wrapper
        with open(main_script, 'rt') as main_module_file:
            main_content = main_module_file.read()
            # try simple wrapper
            m = re.search(r'from (\w+).manage import main', main_content)
            if m:
                project_package_name = m.group(1)
            
            else:
                # pkg_resources kind wrapper
                m = re.search(r"load_entry_point\('(\w+)==", main_content)
                if m:
                    project_package_name = m.group(1)
    
    if project_package_name is None:
        raise Exception('could not identify project package name')  # TODO replace with dedicated conf exception

    _settings_store['PROJECT_PACKAGE_NAME'] = project_package_name

    project_package = importlib.import_module(project_package_name)
    _settings_store.setdefault('PROJECT_PACKAGE_DIR', os.path.dirname(project_package.__file__))


def _setup_default_settings():
    for name, value in inspect.getmembers(defaultsettings):
        if not _is_setting_name(name):
            continue

        _settings_store[name] = value


def _override_project_settings():
    project_settings_module = utils.import_module_or_none('settings')
    if project_settings_module is None:
        # try settings module from project package
        project_settings_path = '{}.settings'.format(_settings_store['PROJECT_PACKAGE_NAME'])
        project_settings_module = utils.import_module_or_none(project_settings_path)
        if project_settings_module is None:
            return

    for name, value in inspect.getmembers(project_settings_module):
        if not _is_setting_name(name):
            continue

        _override_setting(name, value)


def _override_local_settings():
    settings_local_module = utils.import_module_or_none('settingslocal')
    if settings_local_module is None:
        return

    for name, value in inspect.getmembers(settings_local_module):
        if not _is_setting_name(name):
            continue

        _override_setting(name, value)


def _override_env_settings():
    load_dotenv('.env.default')
    load_dotenv('.env', override=True)

    schema_class = settings_schemas.get_all_settings_schema()
    env_vars = schema_class().load(os.environ)

    for name, value in env_vars.items():
        if value is None:
            continue

        _override_setting(name, value)


def _apply_tweaks():
    # update default log level according to DEBUG flag
    if _settings_store['LOGGING'] is defaultsettings.LOGGING and not _settings_store['DEBUG']:
        _settings_store['LOGGING']['root']['level'] = 'INFO'


def _initialize():
    _setup_default_settings()
    _setup_project_package()
    _override_project_settings()
    _override_local_settings()
    _override_env_settings()
    _apply_tweaks()


settings = lazysettings.LazySettings(_settings_store, _initialize)




from colibris import api


def get_object_or_404(model, pk, select_related=None):
    select_related = set(select_related or ())

    try:
        q = model.select(model, *select_related).where(model._meta.primary_key == pk)
        for m in select_related:
            q = q.join(m)

        return q.get()

    except model.DoesNotExist:
        raise api.ModelNotFoundException(model)




from marshmallow import Schema as MMSchema, ValidationError
from marshmallow import pre_dump, post_dump, pre_load, post_load, validates_schema
from marshmallow import fields, validate
from marshmallow import EXCLUDE as MM_EXCLUDE


_settings_schemas = []


class SettingsSchema(MMSchema):
    class Meta:
        unknown = MM_EXCLUDE


class CommonSchema(SettingsSchema):
    DEBUG = fields.Boolean()
    LISTEN = fields.String()
    PORT = fields.Integer()
    MAX_REQUEST_BODY_SIZE = fields.Integer()
    API_DOCS_PATH = fields.String()
    DATABASE = fields.String()


# authentication

class ModelAuthenticationSchema(SettingsSchema):
    AUTHENTICATION_MODEL = fields.String()
    AUTHENTICATION_IDENTITY_FIELD = fields.String()
    AUTHENTICATION_SECRET_FIELD = fields.String()


class JWTAuthenticationSchema(ModelAuthenticationSchema):
    AUTHENTICATION_IDENTITY_CLAIM = fields.String()


class AllAuthenticationSchema(JWTAuthenticationSchema):
    AUTHENTICATION_BACKEND = fields.String()


# authorization

class RoleAuthorizationSchema(SettingsSchema):
    AUTHORIZATION_ROLE_FIELD = fields.String()


class ModelAuthorizationSchema(SettingsSchema):
    AUTHORIZATION_MODEL = fields.String()
    AUTHORIZATION_ACCOUNT_FIELD = fields.String()


class RightsAuthorizationSchema(ModelAuthorizationSchema):
    AUTHORIZATION_RESOURCE_FIELD = fields.String()
    AUTHORIZATION_OPERATIONS_FIELD = fields.String()


class AllAuthorizationSchema(RoleAuthorizationSchema,
                             RightsAuthorizationSchema):

    AUTHORIZATION_BACKEND = fields.String()


# cache

class LocMemCacheSchema(SettingsSchema):
    CACHE_MAX_ENTRIES = fields.Integer()


class RedisCacheSchema(SettingsSchema):
    CACHE_HOST = fields.String()
    CACHE_PORT = fields.Integer()
    CACHE_DB = fields.Integer()
    CACHE_PASSWORD = fields.String()


class AllCacheSchema(LocMemCacheSchema,
                     RedisCacheSchema):

    CACHE_BACKEND = fields.String()


# database

class SQLiteDatabaseSchema(SettingsSchema):
    DATABASE_NAME = fields.String()


class ServerDatabaseSchema(SettingsSchema):
    DATABASE_NAME = fields.String()
    DATABASE_HOST = fields.String()
    DATABASE_PORT = fields.Integer()
    DATABASE_USERNAME = fields.String()
    DATABASE_PASSWORD = fields.String()


class MySQLDatabaseSchema(ServerDatabaseSchema):
    pass


class PostgreSQLDatabaseSchema(ServerDatabaseSchema):
    pass


class AllDatabaseSchema(SQLiteDatabaseSchema,
                        MySQLDatabaseSchema,
                        PostgreSQLDatabaseSchema):

    DATABASE_BACKEND = fields.String()


# cache

class RQTaskQueueSchema(SettingsSchema):
    TASK_QUEUE_HOST = fields.String()
    TASK_QUEUE_PORT = fields.Integer()
    TASK_QUEUE_DB = fields.Integer()
    TASK_QUEUE_PASSWORD = fields.String()
    TASK_QUEUE_POLL_RESULTS_INTERVAL = fields.Integer()


class AllTaskQueueSchema(RQTaskQueueSchema):
    TASK_QUEUE_BACKEND = fields.String()


def register_settings_schema(schema):
    _settings_schemas.append(schema)


def get_all_settings_schema():
    return type('AllSettingsSchema', tuple(_settings_schemas), {})


# put together known schemas

register_settings_schema(CommonSchema)
register_settings_schema(AllAuthenticationSchema)
register_settings_schema(AllAuthorizationSchema)
register_settings_schema(AllCacheSchema)
register_settings_schema(AllDatabaseSchema)
register_settings_schema(AllTaskQueueSchema)




import builtins
import collections
import importlib
import re


def camelcase_to_underscore(s):
    return re.sub('(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))', '_\\1', s).lower().strip('_')


def import_member(path):
    parts = path.split('.')
    module_path = '.'.join(parts[:-1])
    member_name = parts[-1]

    module = importlib.import_module(module_path)

    return getattr(module, member_name)


def import_module_or_none(path):
    ModuleNotFoundError = getattr(builtins, 'ModuleNotFoundError', ImportError)

    try:
        return importlib.import_module(path)

    except ModuleNotFoundError as e:
        if e.name == path:
            return None

        raise


def dict_update_rec(dest, source):
    for k, v in source.items():
        if ((k in dest) and isinstance(dest[k], dict) and
            isinstance(source[k], collections.Mapping)):

            dict_update_rec(dest[k], source[k])

        else:
            dest[k] = source[k]




from colibris.authentication.exceptions import NoSuchAccount, IdentityVerificationFailed


class AuthenticationBackend:
    def extract_auth_data(self, request):
        raise NotImplementedError

    def lookup_account(self, identity):
        raise NotImplementedError

    def verify_identity(self, account, auth_data):
        raise NotImplementedError

    def authenticate(self, request):
        identity, auth_data = self.extract_auth_data(request)
        account = self.lookup_account(identity)
        if account is None:
            raise NoSuchAccount()

        if not self.verify_identity(account, auth_data):
            raise IdentityVerificationFailed()

        return account


class NullBackend(AuthenticationBackend):
    _DUMMY_ACCOUNT = {}

    def __init__(self, **kwargs):
        pass

    def extract_auth_data(self, request):
        return None, None

    def lookup_account(self, identity):
        return self._DUMMY_ACCOUNT

    def verify_identity(self, account, auth_data):
        return True

    def authenticate(self, request):
        return self._DUMMY_ACCOUNT



import argparse
import os
import re
import shutil

TMP_TEMPLATE_PATH = '/tmp/colibris-template-repo'
MAIN_PACKAGE_NAME = '__packagename__'


def start_project():
    parser = argparse.ArgumentParser()
    parser.add_argument('name', help='The project name', type=str)
    parser.add_argument('--template', help='The template repo', type=str, required=False)
    args = parser.parse_args()

    project_name = args.name
    package_name = re.sub('[^a-z0-9_]', '', project_name).lower()

    if args.template is None:
        skeleton_dir = os.path.join(os.path.dirname(__file__), 'skeleton')
    else:
        if os.path.exists(TMP_TEMPLATE_PATH):
            shutil.rmtree(TMP_TEMPLATE_PATH)

        os.system("git clone {} {}".format(args.template, TMP_TEMPLATE_PATH))
        skeleton_dir = TMP_TEMPLATE_PATH

    shutil.copytree(skeleton_dir, project_name, ignore=shutil.ignore_patterns('.git'))

    old_package_name = '{}/{}'.format(project_name, MAIN_PACKAGE_NAME)
    new_package_name = '{}/{}'.format(project_name, package_name)

    shutil.move(old_package_name, new_package_name)

    rename_command = 'find {} -type f | xargs sed -i "s/{}/{}/g"'

    os.system(rename_command.format(project_name, '__packagename__', package_name))
    os.system(rename_command.format(project_name, '__projectname__', project_name))

    print('project {} is ready'.format(project_name))




import time

from colibris.cache.base import CacheBackend, ABSENT


DEFAULT_MAX_ENTRIES = 1e5


class LocMemBackend(CacheBackend):
    def __init__(self, max_entries=DEFAULT_MAX_ENTRIES):
        self._max_entries = max_entries
        self._store = {}

    def get(self, key):
        entry = self._store.get(key, ABSENT)
        if entry is ABSENT:
            return ABSENT

        added, lifetime, value = entry
        if time.time() - added > lifetime:  # entry expired
            self._store.pop(key, None)
            return ABSENT

        return value

    def set(self, key, value, lifetime):
        # clean up expired entries as soon as we hit the limit
        if len(self._store) >= self._max_entries:
            self._cleanup_expired()

        # remove random elements until we have enough room
        while len(self._store) >= self._max_entries:
            self._store.popitem()

        self._store[key] = time.time(), lifetime, value

    def delete(self, key):
        self._store.pop(key, None)

    def _cleanup_expired(self):
        expired_keys = []
        now = time.time()
        for key, (added, lifetime, value) in self._store.items():
            if now - added > lifetime:
                expired_keys.append(key)

        for key in expired_keys:
            self._store.pop(key)




PROJECT_PACKAGE_NAME = 'colibris'

DEBUG = True

LISTEN = '0.0.0.0'
PORT = 8888

MAX_REQUEST_BODY_SIZE = 10 * 1024 * 1024

MIDDLEWARE = [
    'colibris.middleware.handle_errors_json',
    'colibris.middleware.handle_auth',
    'colibris.middleware.handle_schema_validation',
]

AUTHENTICATION = {}
AUTHORIZATION = {}

CACHE = {}
DATABASE = {}
TASK_QUEUE = {}

API_DOCS_PATH = '/api/docs'

LOGGING = {
    'version': 1,
    'formatters': {
        'default': {
            'format': '%(asctime)s: %(levelname)7s: [%(name)s] %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S'
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'default'
        }
    },
    'root': {
        'level': 'DEBUG',
        'handlers': ['console']
    },
    'loggers': {
        'peewee_migrate': {
            'level': 'DEBUG'  # always show details when doing migrations
        }
    }
}

LOGGING_OVERRIDES = {}




class LazySettings:
    def __init__(self, settings_store, initialize):
        self._settings_store = settings_store
        self._initialized = False
        self._initialize = initialize

    def __getattr__(self, name):
        if not self._initialized:
            self._initialize()
            self._initialized = True

        try:
            return self._settings_store[name]

        except KeyError as e:
            # we want attribute errors instead of key errors on settings
            raise AttributeError(e)




import json
import logging
import re

from aiohttp import web
from aiohttp_apispec import validation_middleware
from webargs import aiohttpparser

from colibris import api
from colibris import app
from colibris import authentication
from colibris import settings
from colibris import utils
from colibris.api import BaseJSONException, envelope


logger = logging.getLogger(__name__)

_authentication_backend_settings = dict(settings.AUTHENTICATION)
_authentication_backend_path = _authentication_backend_settings.pop('backend',
                                                                    'colibris.authentication.base.NullBackend')
_authentication_backend_class = utils.import_member(_authentication_backend_path)
_authentication_backend = _authentication_backend_class(**_authentication_backend_settings)

_authorization_backend_settings = dict(settings.AUTHORIZATION)
_authorization_backend_path = _authorization_backend_settings.pop('backend',
                                                                  'colibris.authorization.base.NullBackend')
_authorization_backend_class = utils.import_member(_authorization_backend_path)
_authorization_backend = _authorization_backend_class(**_authorization_backend_settings)


class HTTPSchemaValidationError(web.HTTPUnprocessableEntity):
    def __init__(self, schema_error, **kwargs):
        super().__init__(**kwargs)

        self.details = schema_error.messages


@web.middleware
async def handle_errors_json(request, handler):
    try:
        return await handler(request)

    except BaseJSONException as e:
        return web.json_response(envelope.wrap_error(e.code, e.message, e.details), status=e.status)

    except HTTPSchemaValidationError as e:
        code = 'invalid_fields'
        message = 'Some of the supplied fields are invalid.'

        return web.json_response(envelope.wrap_error(code, message, e.details), status=e.status)

    except web.HTTPRedirection as e:
        return e

    except (web.HTTPClientError, web.HTTPServerError) as e:
        code = utils.camelcase_to_underscore(re.sub('[^a-zA-Z0-9_]', '', e.reason))
        message = e.reason

        return web.json_response(envelope.wrap_error(code, message), status=e.status)

    except json.JSONDecodeError as e:
        code = 'invalid_json'
        message = 'Invalid JSON.'
        details = str(e)

        return web.json_response(envelope.wrap_error(code, message, details), status=400)

    except app.HealthException as e:
        code = 'unhealthy'
        message = 'Service is not healthy.'
        details = str(e)

        return web.json_response(envelope.wrap_error(code, message, details), status=500)

    except Exception as e:
        code = 'server_error'
        message = 'Internal Server Error'
        details = None

        if settings.DEBUG:
            import traceback

            logger.error('internal server error: %s', e, exc_info=True)
            details = traceback.format_exc()

        return web.json_response(envelope.wrap_error(code, message, details), status=500)


@web.middleware
async def handle_auth(request, handler):
    if request.match_info.http_exception is not None:
        raise request.match_info.http_exception

    path = request.match_info.route.resource.canonical
    route = app.routes_by_path.get(path)
    if not route:  # shouldn't happen
        raise web.HTTPNotFound()

    method, path, _handler, permissions = route

    # only go through authentication if route specifies permissions;
    # otherwise route is considered public
    if permissions:
        try:
            account = _authentication_backend.authenticate(request)

        except authentication.AuthenticationException as e:
            logger.error('%s %s authentication failed: %s', method, path, e)

            raise api.UnauthenticatedException()

        # at this point we can safely associate request with account
        request.account = account

        if not _authorization_backend.authorize(account, method, path, permissions):
            logger.error('%s %s forbidden for %s', method, path, account)

            raise api.ForbiddenException()

    return await handler(request)


@web.middleware
async def handle_schema_validation(request, handler):
    return await validation_middleware(request, handler)


@aiohttpparser.parser.error_handler
def _handle_schema_validation_error(error, req, schema):
    raise HTTPSchemaValidationError(error)
